"use strict";

var _iniettore = _interopRequireWildcard(require("../../src/iniettore"));

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = Object.defineProperty && Object.getOwnPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : {}; if (desc.get || desc.set) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } } newObj.default = obj; return newObj; } }

describe('Given a TRANSIENT, SINGLETON, CONSTRUCTOR mapping', () => {
  describe('when requesting it', () => {
    it('should create an instance at the first request and return it at the subsequent call', () => {
      var constructorSpy = jest.fn();
      var bar1, bar2;

      class Bar {
        constructor() {
          constructorSpy.call();
        }

      }

      var rootContext = _iniettore.default.create(map => {
        map('bar').to(Bar).as(_iniettore.TRANSIENT, _iniettore.SINGLETON, _iniettore.CONSTRUCTOR);
      });

      expect(constructorSpy).not.toHaveBeenCalled();
      bar1 = rootContext.get('bar');
      expect(constructorSpy).toHaveBeenCalledTimes(1);
      bar2 = rootContext.get('bar');
      expect(bar2).toBe(bar1);
    });
  });
  describe('with TRANSIENT, SINGLETON, CONSTRUCTOR as shared dependency', () => {
    describe('when requesting it', () => {
      it('should create singleton instances from the respective constructors', () => {
        var commonInstance;

        class Common {
          constructor() {
            commonInstance = this;
          }

        }

        class Bar {
          constructor(common) {
            expect(common).toBeInstanceOf(Common);
            expect(common).toBe(commonInstance);
          }

        }

        class Foo {
          constructor(common, bar) {
            expect(common).toBeInstanceOf(Common);
            expect(common).toBe(commonInstance);
            expect(bar).toBeInstanceOf(Bar);
          }

        }

        var rootContext = _iniettore.default.create(map => {
          map('common').to(Common).as(_iniettore.TRANSIENT, _iniettore.SINGLETON, _iniettore.CONSTRUCTOR);
          map('bar').to(Bar).as(_iniettore.TRANSIENT, _iniettore.SINGLETON, _iniettore.CONSTRUCTOR).injecting('common');
          map('foo').to(Foo).as(_iniettore.TRANSIENT, _iniettore.SINGLETON, _iniettore.CONSTRUCTOR).injecting('common', 'bar');
        });

        expect(rootContext.get('foo')).toBeInstanceOf(Foo);
      });
    });
  });
  describe('and an instance of it', () => {
    var rootContext;
    describe('when releasing the instance', () => {
      class Bar {}

      beforeEach(() => {
        rootContext = _iniettore.default.create(map => {
          map('bar').to(Bar).as(_iniettore.TRANSIENT, _iniettore.SINGLETON, _iniettore.CONSTRUCTOR);
        });
      });
      it('should not release the instance if still in use', () => {
        var bar1 = rootContext.get('bar');
        var bar2 = rootContext.get('bar');
        expect(bar1).toBe(bar2);
        rootContext.release('bar');
        let bar3 = rootContext.get('bar');
        expect(bar3).toBe(bar1);
      });
      describe('the same amount of times it has been requested', () => {
        it('should release the instance', () => {
          var bar1 = rootContext.get('bar');
          var bar2 = rootContext.get('bar');
          expect(bar1).toBe(bar2);
          rootContext.release('bar');
          rootContext.release('bar');
          let bar3 = rootContext.get('bar');
          expect(bar3).not.toBe(bar1);
        });
      });
    });
  });
});
describe('Given a TRANSIENT, SINGLETON, PROVIDER mapping', () => {
  describe('when requesting it', () => {
    it('should invoke the provider function at the first request and return the created item at the subsequent call', () => {
      var providerSpy = jest.fn();

      var fooProvider = () => {
        providerSpy();
        return {};
      };

      var foo1, foo2;

      var rootContext = _iniettore.default.create(map => {
        map('foo').to(fooProvider).as(_iniettore.TRANSIENT, _iniettore.SINGLETON, _iniettore.PROVIDER);
      });

      expect(providerSpy).not.toHaveBeenCalled();
      foo1 = rootContext.get('foo');
      expect(providerSpy).toHaveBeenCalledTimes(1);
      foo2 = rootContext.get('foo');
      expect(foo2).toBe(foo1);
    });
  });
  describe('and an instance created from it', () => {
    var rootContext;
    describe('when releasing the instance', () => {
      function fooProvider() {
        return {};
      }

      beforeEach(() => {
        rootContext = _iniettore.default.create(map => {
          map('foo').to(fooProvider).as(_iniettore.TRANSIENT, _iniettore.SINGLETON, _iniettore.PROVIDER);
        });
      });
      it('should not release the instance if still in use', () => {
        var foo1 = rootContext.get('foo');
        var foo2 = rootContext.get('foo');
        expect(foo1).toBe(foo2);
        rootContext.release('foo');
        let foo3 = rootContext.get('foo');
        expect(foo3).toBe(foo1);
      });
      describe('the same amount of times it has been requested', () => {
        it('should release the instance', () => {
          var foo1 = rootContext.get('foo');
          var foo2 = rootContext.get('foo');
          expect(foo1).toBe(foo2);
          rootContext.release('foo');
          rootContext.release('foo');
          let foo3 = rootContext.get('foo');
          expect(foo3).not.toBe(foo1);
        });
      });
    });
  });
});